/**
 * $Id: ClasspathMonitor.java,v 1.40 2009/08/21 19:59:22 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 10.02.2009 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.monitor;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.jar.JarFile;
import java.util.zip.*;

import javax.management.JMException;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.impl.LogFactoryImpl;

import patterntesting.runtime.annotation.*;
import patterntesting.runtime.jmx.MBeanHelper;
import patterntesting.runtime.util.*;

/**
 * To avoid classpath problems like double entries of the same class or resource
 * in the classpath there are several methods available. <br/>
 * To get the boot classpath the sytem property "sun.boot.class.path" is used to
 * get them. This will work of course only for the SunVM.
 * 
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 10.02.2009
 * @version $Revision: 1.40 $
 * @see http://www.javaworld.com/javaworld/javatips/jw-javatip105.html
 */
public class ClasspathMonitor extends Thread implements ClasspathMonitorMBean {

    private static final long serialVersionUID = 20090511L;
    private static final Log log = LogFactoryImpl
            .getLog(ClasspathMonitor.class);
    protected final ClasspathDigger classpathDigger;
    private final ClassLoader cloader;
    private final String[] classpath;
    private String[] loadedClasses = new String[0];
    private final Set<String> unusedClasses;
    private final String[] classpathClasses;
    protected List<Class<?>> doubletList = new ArrayList<Class<?>>();
    private List<Class<?>> incompatibleClassList = new ArrayList<Class<?>>();
    private static ClasspathMonitor instance;

    static {
        instance = new ClasspathMonitor();
    }

    protected ClasspathMonitor() {
        this.classpathDigger = new ClasspathDigger();
        this.cloader = this.classpathDigger.getClassLoader();
        this.classpath = this.classpathDigger.getClasspath();
        this.unusedClasses = getClasspathClassSet();
        this.classpathClasses = createClasspathClasses();
    }

    /**
     * At the beginning unusedClasses contains all classes found in the
     * classpath. So we use this attribute to get the classpathClasses.
     */
    @ProfileMe
    private String[] createClasspathClasses() {
        Set<String> classSet = unusedClasses;
        String[] classpathClasses = new String[classSet.size()];
        classSet.toArray(classpathClasses);
        return classpathClasses;
    }

    /**
     * Yes, it is a Singleton because it offers only some services. So we don't
     * need the object twice.
     * 
     * @return the only instance
     */
    public static ClasspathMonitor getInstance() {
        return instance;
    }

    public static void registerAsMBean() {
        try {
            MBeanHelper.registerMBean(getInstance());
        } catch (JMException e) {
            if (log.isInfoEnabled()) {
                log.info(instance + " can't be registered as MBean (" + e
                		+ ")");
            }
        }
    }

    /**
     * @see ClasspathMonitorMBean#whichResource(String)
     */
    @ProfileMe
    public URI whichResource(String name) {
        return ClasspathDigger.whichResource(name, this.cloader);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#whichClass(java.lang.String)
     */
    @ProfileMe
    public URI whichClass(String name) {
        String resource = Converter.classToResource(name);
        return whichResource(resource);
    }

    /**
     * Where is the given class found in the classpath? Which class was loaded
     * from which URI?
     * 
     * @param clazz
     *            the class
     * @return file or jar path
     */
    @ProfileMe
    public URI whichClass(Class<?> clazz) {
        return whichClass(clazz.getName());
    }

    /**
     * Returns the jar file or path where the given classname was found.
     * 
     * @param classname
     *            e.g. java.lang.String
     * @return jar or path as URI
     */
    public URI whichClassPath(String classname) {
        String resource = Converter.classToResource(classname);
        return whichResourcePath(resource);
    }

    /**
     * Returns the jar file or path where the given class was found.
     * 
     * @param clazz
     *            e.g. Sting.class
     * @return jar or path as URI
     */
    public URI whichClassPath(Class<?> clazz) {
        return whichClassPath(clazz.getName());
    }
    
    public URI whichClassPath(Package p) {
        String resource = Converter.toResource(p);
        return whichResourcePath(resource);        
    }

    /**
     * Returns the jar file or path where the given resource was found
     * 
     * @param resource
     *            e.g. log4j.properties
     * @return jar or path as URI
     */
    public URI whichResourcePath(String resource) {
        URI uri = this.whichResource(resource);
        if (uri == null) {
            return null;
        }
        return ClasspathHelper.getParent(uri, resource);
    }

    public JarFile whichClassJar(Class<?> clazz) {
        return whichClassJar(clazz.getName());
    }

    public JarFile whichClassJar(String classname) {
        String resource = Converter.classToResource(classname);
        return whichResourceJar(resource);
    }

    public JarFile whichResourceJar(String resource) {
        return whichResourceJar(this.whichResourcePath(resource));
    }

    public static JarFile whichResourceJar(URI resource) {
        if (resource == null) {
            return null;
        }
        File file = Converter.toFile(resource);
        try {
            return new JarFile(file);
        } catch (IOException ioe) {
            log.debug("can't read " + file, ioe);
            return null;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getResources(java.lang.String
     * )
     */
    @ProfileMe
    public Enumeration<URL> getResources(String name) {
        try {
            Enumeration<URL> resources = cloader.getResources(name);
            if (!resources.hasMoreElements()) {
                if (name.startsWith("/")) {
                    return getResources(name.substring(1));
                }
                if (log.isDebugEnabled()) {
                    log.debug(name + " not found in classpath");
                }
            }
            return resources;
        } catch (IOException ioe) {
            log.info(name + " not found in classpath", ioe);
            return null;
        }
    }

    public static URL getResource(String name) {
        return ClasspathMonitor.class.getResource(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getNoResources(java.lang
     * .String)
     */
    @ProfileMe
    public int getNoResources(String name) {
        Enumeration<URL> resources = getResources(name);
        if (resources == null) {
            return 0;
        }
        int n = 0;
        while (resources.hasMoreElements()) {
            n++;
            resources.nextElement();
        }
        if (log.isTraceEnabled()) {
            log.trace(n + " element(s) of " + name + " found in classpath");
        }
        return n;
    }

    public int getNoClasses(Class<?> cl) {
        return getNoClasses(cl.getName());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getNoClasses(java.lang.String
     * )
     */
    public int getNoClasses(String classname) {
        return getNoResources(Converter.classToResource(classname));
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#isDoublet(java.lang.String)
     */
    @ProfileMe
    public boolean isDoublet(String name) {
        Enumeration<URL> resources = getResources(name);
        if ((resources == null) || !resources.hasMoreElements()) {
            throw new NoSuchElementException(name);
        }
        resources.nextElement();
        if (resources.hasMoreElements()) {
            logDoublets(name);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Is the given class a doublet, i.e. can it be found more than once in the
     * classpath?
     * 
     * @param clazz
     *            the class to be checked
     * @return true if class is found more than once in the classpath
     * @throws NoSuchElementException
     *             if class is not found in the classpath
     */
    @ProfileMe
    public boolean isDoublet(Class<?> clazz) {
        String classname = clazz.getName();
        String resource = Converter.classToResource(classname);
        return isDoublet(resource);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#getFirstDoublet(java.lang
     *      .String)
     */
    public URL getFirstDoublet(String name) {
        return getDoublet(name, 1);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#getDoublet(java.lang.String,
     *      int)
     */
    public URL getDoublet(String name, int nr) {
        Enumeration<URL> resources = getResources(name);
        for (int i = 0; resources.hasMoreElements(); i++) {
            URL url = (URL) resources.nextElement();
            if (i == nr) {
                return url;
            }
        }
        return null;
    }

    public URL getDoublet(Class<?> clazz, int nr) {
        String resource = Converter.classToResource(clazz.getName());
        return getDoublet(resource, nr);
    }

    private void logDoublets(String name) {
        if (log.isTraceEnabled()) {
            List<URL> doublets = new Vector<URL>();
            Enumeration<URL> resources = getResources(name);
            while (resources.hasMoreElements()) {
                URL url = resources.nextElement();
                doublets.add(url);
            }
            log.trace(name + " doublets found: " + doublets);
        }
    }

    @ProfileMe
    public synchronized List<Class<?>> getDoubletList() {
        if (multiThreadingEnabled) {
            return getDoubletListParallel();
        } else {
            return getDoubletListSerial();
        }
    }

    /**
     * Looks for each loaded class if it is a doublet or not. For the
     * performance reason it looks in the doubletList from the last time if it
     * is already found. This is done because normally the number of doublets
     * does not decrease.
     * 
     * @return a sorted list of found doublets
     */
    @ProfileMe
    protected synchronized List<Class<?>> getDoubletListSerial() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        for (Class<?> clazz : loadedClassList) {
            // TODO this can be parallized
            if (doubletList.contains(clazz)) {
                continue;
            }
            try {
                if (this.isDoublet(clazz)) {
                    doubletList.add(clazz);
                }
            } catch (NoSuchElementException nsee) {
                if (log.isTraceEnabled()) {
                    log.trace(clazz + " not found -> ignored");
                }
            }
        }
        Collections.sort(doubletList, new ObjectComparator());
        return doubletList;
    }
    
    /**
     * @see ClasspathMonitorMBean#getDoublets()
     */
    @ProfileMe
    public String[] getDoublets() {
        log.debug("calculating doublets...");
        List<Class<?>> doubletList = this.getDoubletList();
        String[] doublets = new String[doubletList.size()];
        for (int i = 0; i < doublets.length; i++) {
            doublets[i] = doubletList.get(i).toString();
        }
        return doublets;
    }

    /**
     * @see ClasspathMonitorMBean#getDoubletClasspath()
     */
    @ProfileMe
    public String[] getDoubletClasspath() {
        log.debug("calculating doublet-classpath...");
        Set<URI> classpathSet = getClasspathSet(this.getDoubletList());
        return getAsArray(classpathSet);
    }

    private SortedSet<URI> getClasspathSet(List<Class<?>> classes) {
        SortedSet<URI> classpath = new TreeSet<URI>();
        for (Class<?> cl : classes) {
            String resource = Converter.toResource(cl);
            Enumeration<URL> resources = this.getResources(resource);
            while (resources.hasMoreElements()) {
                URL url = resources.nextElement();
                try {
                    URI path = url.toURI();
                    classpath.add(ClasspathHelper.getParent(path, resource));
                } catch (URISyntaxException canthappen) {
                    log.debug("can't convert " + url + " to URI", canthappen);
                }
            }
        }
        return classpath;
    }

    /**
     * Dumps the internal fields of the classloader (after an idea from
     * "Java ist auch eine Insel").
     * 
     * @see http 
     *      ://www.galileocomputing.de/openbook/javainsel6/javainsel_21_003.htm
     *      #mj7e2d89f4b96fde1900bc09fd1db83fb1
     */
    @ProfileMe
    public String getClassLoaderDetails() {
        StringBuffer sbuf = new StringBuffer("dump of " + cloader + ":\n");
        for (Class<?> cl = this.cloader.getClass(); cl != null; cl = cl
                .getSuperclass()) {
            sbuf.append('\t');
            dumpFields(sbuf, cl, this.cloader);
        }
        return sbuf.toString().trim();
    }

    @ProfileMe
    private static final void dumpFields(StringBuffer sbuf, Class<?> cl,
            Object obj) {
        Field[] fields = cl.getDeclaredFields();
        AccessibleObject.setAccessible(fields, true);
        for (int i = 0; i < fields.length; i++) {
            sbuf.append(fields[i]);
            sbuf.append(" = ");
            try {
                sbuf.append(fields[i].get(obj));
            } catch (Exception e) {
                sbuf.append("<" + e + ">");
            }
            sbuf.append('\n');
        }
    }

    /**
     * @return true if it is a known classloader
     */
    public boolean isClassloaderSupported() {
        return this.classpathDigger.isClassloaderSupported();
    }

    /**
     * Returns some information about the classloader. At least the user should
     * be informed if it is a unknown classloader which is not supported or not
     * tested.
     * 
     * @return e.g. "unknown classloader xxx - classpath can be wrong"
     */
    public String getClassloaderInfo() {
        String info = this.isClassloaderSupported() ? "supported" : "unsupported";
        return this.cloader.getClass().getName() + " (" + info + ")";
    }

    /**
     * @see ClasspathMonitorMBean#getLoadedPackages()
     */
    public String[] getLoadedPackages() {
        Package[] packages = this.classpathDigger.getLoadedPackageArray();
        String[] strings = new String[packages.length];
        for (int i = 0; i < packages.length; i++) {
            strings[i] = packages[i].toString();
        }
        Arrays.sort(strings);
        return strings;
    }
    
    /**
     * This method is used by PatternTesting Samples (in packages.jsp). So it
     * is left here as wrapper around ClasspathDigger.
     * 
     * @return array of packages
     */
    public Package[] getLoadedPackageArray() {
    	return this.classpathDigger.getLoadedPackageArray();
    }

    /**
     * If you want to dump all packages you can use this method. The output will
     * be sorted.
     * 
     * @return each package in a single line
     */
    public String getLoadedPackagesAsString() {
        String[] packages = getLoadedPackages();
        StringBuffer sbuf = new StringBuffer();
        for (int i = 0; i < packages.length; i++) {
            sbuf.append(packages[i]);
            sbuf.append('\n');
        }
        return sbuf.toString().trim();
    }

    /**
     * Returns a list of classes which were loaded by the given classloader. <br/>
     * Ok, we must do some hacks here: there is an undocumented attribute
     * "classes" which contains the loaded classes. <br/>
     * HANDLE WITH CARE (it's a hack and it depends on the used classloader)
     * 
     * @param cloader
     *            classloader
     * @return list of classes
     */
    @ProfileMe
    @SuppressWarnings("unchecked")
    private static List<Class<?>> getLoadedClassList(ClassLoader cloader) {
        try {
            Field field = ReflectionHelper.getField(cloader.getClass(),
                    "classes");
            List<Class<?>> classList = (List<Class<?>>) field.get(cloader);
            return new ArrayList<Class<?>>(classList);
        } catch (Exception e) {
            log.warn("can't access field 'classes'", e);
            return Collections.EMPTY_LIST;
        }
    }

    /**
     * Returns a list of classes which were loaded by the classloader.
     * 
     * @return list of classes
     */
    @ProfileMe
    synchronized public List<Class<?>> getLoadedClassList() {
        List<Class<?>> classlist = getLoadedClassList(cloader);
        return classlist;
    }

    /**
     * As MBean a string array could be displayed by the 'jconsole'. A class
     * array not.
     * 
     * @return the classes as sorted string array
     */
    synchronized public String[] getLoadedClasses() {
        List<Class<?>> classes = getLoadedClassList();
        if (classes.size() != loadedClasses.length) {
            loadedClasses = new String[classes.size()];
            for (int i = 0; i < loadedClasses.length; i++) {
                loadedClasses[i] = classes.get(i).toString();
            }
            Arrays.sort(loadedClasses);
        }
        return loadedClasses;
    }

    @ProfileMe
    public String getLoadedClassesAsString() {
        List<Class<?>> classes = getLoadedClassList();
        Collections.sort(classes, new ObjectComparator());
        StringBuffer sbuf = new StringBuffer();
        for (Iterator<Class<?>> i = classes.iterator(); i.hasNext();) {
            sbuf.append(i.next().toString().trim());
            sbuf.append('\n');
        }
        return sbuf.toString().trim();
    }

    public boolean isLoaded(String classname) {
        for (Iterator<Class<?>> iterator = getLoadedClassList().iterator(); iterator
                .hasNext();) {
            Class<?> loaded = (Class<?>) iterator.next();
            if (classname.equals(loaded.getName())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Unused classes are classes which are not loaded but which are found in
     * the classpath.
     * 
     * @return unused classes
     * @see #getLoadedClasses()
     * @see #getClasspathClasses()
     */
    @ProfileMe
    public String[] getUnusedClasses() {
        List<Class<?>> classes = getLoadedClassList();
        Collection<String> used = new ArrayList<String>();
        for (Class<?> cl : classes) {
            String classname = cl.getName();
            if (unusedClasses.contains(classname)) {
                used.add(classname);
            }
        }
        unusedClasses.removeAll(used);
        String[] unused = new String[unusedClasses.size()];
        unusedClasses.toArray(unused);
        return unused;
    }

    /**
     * Scans the classpath for all classes.
     * 
     * @return all classes as String array
     */
    public String[] getClasspathClasses() {
        return classpathClasses;
    }

    @ProfileMe
    private Set<String> getClasspathClassSet() {
        Set<String> classSet = new TreeSet<String>();
        for (int i = 0; i < classpath.length; i++) {
            addClasses(classSet, new File(classpath[i]));
        }
        return classSet;
    }

    private static void addClasses(Set<String> classSet, File path) {
        if (log.isTraceEnabled()) {
            log.trace("adding classes from " + path.getAbsolutePath() + "...");
        }
        try {
            if (path.isDirectory()) {
                addClassesFromDir(classSet, path);
            } else {
                addClassesFromArchive(classSet, path);
            }
        } catch (IOException e) {
            log.warn("can't add classes from " + path.getAbsolutePath());
        }
    }

    @ProfileMe
    private static void addClassesFromDir(Set<String> classSet, File dir)
            throws IOException {
        ClassWalker classWalker = new ClassWalker(dir);
        Collection<String> classes = classWalker.getClasses();
        classSet.addAll(classes);
    }

    @ProfileMe
    private static void addClassesFromArchive(Set<String> classSet, File archive)
            throws IOException {
        ZipFile zipFile = new ZipFile(archive);
        Enumeration<? extends ZipEntry> entries = zipFile.entries();
        while (entries.hasMoreElements()) {
            ZipEntry entry = entries.nextElement();
            String name = entry.getName();
            if (name.endsWith(".class")) {
                classSet.add(Converter.resourceToClass(name));
            }
        }
    }

    /**
     * Gets the loaded classpath (without the bootclasspath) as sorted set. <br/>
     * TODO: this method lets place for more performance (e.g. using a
     * Hashtable<Class, URI> for caching)
     * 
     * @return the loaded classpath (excluding the bootclasspath)
     */
    @ProfileMe
    public SortedSet<URI> getUsedClasspathSet() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        SortedSet<URI> usedClasspathSet = new TreeSet<URI>();
        for (Class<?> clazz : loadedClassList) {
            URI classUri = this.whichClass(clazz);
            if (classUri != null) {
                URI classpath = ClasspathHelper.getParent(classUri, clazz);
                usedClasspathSet.add(classpath);
            }
        }
        return usedClasspathSet;
    }

    /**
     * It might be that the used classpath changes during the calculation of it.
     * So it is only a good approximation how it looks at the moment of the
     * call.
     * 
     * @see ClasspathMonitor#getUsedClasspath()
     */
    @ProfileMe
    public String[] getUsedClasspath() {
        log.debug("calculating used classpath...");
        SortedSet<URI> classpathSet = this.getUsedClasspathSet();
        return getAsArray(classpathSet);
    }

    private static String[] getAsArray(Set<URI> classpathSet) {
        String[] classpath = new String[classpathSet.size()];
        Object[] a = classpathSet.toArray();
        for (int i = 0; i < classpath.length; i++) {
            URI uri = (URI) a[i];
            classpath[i] = uri.getPath();
            if (StringUtils.isEmpty(classpath[i])) {
                classpath[i] = StringUtils.substringAfterLast(uri.toString(),
                        ":");
            }
            if (classpath[i].endsWith(File.separator)) {
                classpath[i] = classpath[i].substring(0,
                        (classpath[i].length() - 1));
            }
        }
        Arrays.sort(classpath);
        return classpath;
    }

    /**
     * The unused classpath is this path which are not used in past. <br/>
     * TODO: the first three calls can be parallized (for better performance)
     * 
     * @return the unused classpath (sorted)
     */
    @ProfileMe
    public String[] getUnusedClasspath() {
        log.debug("calculating unused classpath...");
        SortedSet<String> unused = new TreeSet<String>();
        String[] classpath = this.getClasspath();
        log.trace(Arrays.class + " loaded (to get corrected used classpath");
        String[] used = this.getUsedClasspath();
        for (int i = 0; i < classpath.length; i++) {
            String path = classpath[i];
            if (Arrays.binarySearch(used, path) < 0) {
                unused.add(path);
            }
        }
        String[] a = new String[unused.size()];
        return unused.toArray(a);
    }

    /**
     * To get the boot classpath the sytem property "sun.boot.class.path" is
     * used to get them. This will work of course only for the SunVM.
     * 
     * @return the boot classpath as String array
     */
    public String[] getBootClasspath() {
        return this.classpathDigger.getBootClasspath();
    }

    /**
     * The classpath is stored in the system property "java.class.path". And
     * this is the classpath which will be returned.
     * 
     * @return the classpath as String array
     */
    public String[] getClasspath() {
        return this.classpath;
    }

    public Long getSerialVersionUID(String classname)
            throws IllegalAccessException {
        try {
            Class<?> clazz = Class.forName(classname);
            return getSerialVersionUID(clazz);
        } catch (ClassNotFoundException e) {
            log.info(classname + " not found (" + e.getLocalizedMessage()
            		+ ")");
            return null;
        }
    }

    public Long getSerialVersionUID(Class<?> clazz)
            throws IllegalAccessException {
        try {
            Field field = ReflectionHelper.getField(clazz, "serialVersionUID");
            return (Long) field.get(null);
        } catch (NoSuchFieldException e) {
            log.debug(clazz + " has no serialVersionUID");
            return null;
        }
    }

    /**
     * Looks for each doublet if it has a different doublets. For the
     * performance reason it looks in the incompatibleClassList from the last
     * time if it is already found. This is done because normally the number of
     * incompatible classed does not decrease.
     * 
     * @return a sorted list of incompatible classes
     */
    @ProfileMe
    public synchronized List<Class<?>> getIncompatibleClassList() {
        List<Class<?>> doublets = this.getDoubletList();
        for (Class<?> clazz : doublets) {
            if (incompatibleClassList.contains(clazz)) {
                continue;
            }
            String resource = Converter.classToResource(clazz.getName());
            Enumeration<URL> resources = getResources(resource);
            URL url = (URL) resources.nextElement();
            ArchivEntry archivEntry = new ArchivEntry(url);
            while (resources.hasMoreElements()) {
                url = (URL) resources.nextElement();
                ArchivEntry doubletEntry = new ArchivEntry(url);
                if (archivEntry.equals(doubletEntry)) {
                    incompatibleClassList.add(clazz);
                    break;
                }
            }
        }
        return incompatibleClassList;
    }

    /**
     * Incompatible classes are doublets with different byte codes.
     * 
     * @return doublet classes with different byte codes.
     */
    public String[] getIncompatibleClasses() {
        log.debug("calculating incompatible classes...");
        List<Class<?>> classList = this.getIncompatibleClassList();
        String[] classes = new String[classList.size()];
        for (int i = 0; i < classes.length; i++) {
            classes[i] = classList.get(i).toString();
        }
        return classes;
    }

    /**
     * @return the classpathes where incompatible classes were found
     */
    public String[] getIncompatibleClasspath() {
        log.debug("calculating doublet-classpath...");
        Set<URI> classpathSet = getClasspathSet(this.getIncompatibleClassList());
        return getAsArray(classpathSet);
    }

    /**
     * You can register the instance as shutdown hook. If the VM is terminated
     * the attributes are logged and dumped to a text file in the tmp directory.
     * 
     * @see #logStatistic()
     * @see #dumpStatistic()
     */
    public static void addAsShutdownHook() {
        Runtime.getRuntime().addShutdownHook(instance);
        if (log.isDebugEnabled()) {
            log.debug(instance + " registered as shutdown hook");
        }
    }

    /**
     * This method is called when the ClasspathMonitor is registered as shutdown
     * hook.
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        dumpMe();
    }

    /**
     * Logs the different array to the log output.
     */
    public void logMe() {
        try {
            StringWriter writer = new StringWriter();
            this.dumpMe(writer);
            log.info(writer.toString());
        } catch (IOException e) {
            log.info(this.getClassloaderInfo());
        }
    }

    /**
     * This operation dumps the different MBean attributes to a temporary file
     * with the prefix "cpmon" (for ClasspathMonitor) and the extension ".txt".
     */
    public void dumpMe() {
        try {
            File dumpFile = File.createTempFile("cpmon", ".txt");
            Writer writer = new FileWriter(dumpFile);
            dumpMe(writer);
            writer.close();
            log.info("ClasspathMonitor attributes dumped to " + dumpFile);
        } catch (IOException ioe) {
            log.info("can't dump ClasspathMonitor attributes (" + ioe + ")");
        }
    }

    private void dumpMe(Writer unbuffered) throws IOException {
        BufferedWriter writer = new BufferedWriter(unbuffered);
        dumpArray(writer, "BootClasspath", getBootClasspath());
        dumpArray(writer, "Classpath", getClasspath());
        dumpArray(writer, "ClasspathClasses", getClasspathClasses());
        dumpArray(writer, "DoubletClasspath", getDoubletClasspath());
        dumpArray(writer, "Doublets", getDoublets());
        dumpArray(writer, "LoadedClasses", getLoadedClasses());
        dumpArray(writer, "LoadedPackages", getLoadedPackages());
        dumpArray(writer, "IncompatibleClasses", getIncompatibleClasses());
        dumpArray(writer, "IncompatibleClasspath", getIncompatibleClasspath());
        dumpArray(writer, "UnusedClasses", getUnusedClasses());
        dumpArray(writer, "UnusedClasspath", getUnusedClasspath());
        dumpArray(writer, "UsedClasspath", getUsedClasspath());
        dumpClassloaderInfo(writer);
        writer.flush();
    }
    
    private void dumpClassloaderInfo(BufferedWriter writer) throws IOException {
        writer.write("=== ClassloaderInfo ===");
        writer.newLine();
        writer.write(this.getClassloaderInfo());
        writer.newLine();
        writer.newLine();
        writer.write("----- ClassLoaderDetails -----");
        writer.newLine();
        writer.write(this.getClassLoaderDetails());
        writer.newLine();
        writer.newLine();
    }

    private static void dumpArray(BufferedWriter writer, String title,
            String[] array) throws IOException {
        writer.write("=== ");
        writer.write(title);
        writer.write(" ===");
        writer.newLine();
        for (int i = 0; i < array.length; i++) {
            writer.write(array[i]);
            writer.newLine();
        }
        writer.newLine();
    }

    /**
     * We don't want to use the toString() implementation of the Thread class so
     * we use our own one.
     * 
     * @see java.lang.Thread#toString()
     */
    @Override
    public String toString() {
        return this.getClass().getSimpleName() + " for " + this.cloader;
    }

    /**
     * This main method is only for testing the ClasspathMonitor together with
     * the 'jconsole'. On a Java5 VM start it with the following options:
     * -Dcom.sun.management.jmxremote.local.only=false
     * -Dcom.sun.management.jmxremote
     * 
     * @throws JMException
     */
    public static void main(String[] args) throws JMException {
        ClasspathMonitor.registerAsMBean();
        ThreadUtil.sleep(300, TimeUnit.SECONDS);
    }

    
    
    /////   M U L T I   T H R E A D I N G   S E C T I O N   ///////////////////
    
    private boolean multiThreadingEnabled = getMultiThreadingEnabled();
    
    /**
     * Multi threading is enabled if more than one processor will be found
     * or if property "multiThreadingEnabled=true" is set.
     * 
     * @return true on multi core processors
     */
    private static boolean getMultiThreadingEnabled() {
        String s = System.getProperty("multiThreadingEnabled");
        if (s != null) {
            if (log.isDebugEnabled()) {
                log.debug("multiThreadingEnabled=" + s);
            }
            return s.equalsIgnoreCase("true");
        }
        int n = Runtime.getRuntime().availableProcessors();
        boolean enabled = (n > 1);
        if (log.isDebugEnabled()) {
            log.debug(n + " processors found, multi threading "
                    + (enabled ? "enabled" : "not enabled"));
        }
        return enabled;
    }
    
    /**
     * Is multi threading enabled?
     * Multi threading is automatically enabled if more than one processor
     * is found. Otherwise you can use set the system property
     * "multiThreadingEnabled=true" to activate it.
     * 
     * @since 0.9.7
     * 
     * @return true if multi threading is enabled for this class.
     */
    public final boolean isMultiThreadingEnabled() {
        return multiThreadingEnabled;
    }
    
    /**
     * Here you can enable or disable the (experimetal) multi threading mode to
     * see if it is really faster on a mult-core machine.
     * 
     * @since 0.9.7
     * 
     * @param enabled
     */
    public final void setMultiThreadingEnabled(boolean enabled) {
        multiThreadingEnabled = enabled;
    }
    
    /**
     * This is an experiment: can we tune getDoubleList() by using thread
     * techniques like consumer-/producer-pattern? Or is the synchronization
     * overhead to big? Let's start and compare later.
     * <br/>
     * Multi threading is automatically activated if property
     * "multiThreadingEnabled" is set to true or if more than one available
     * processor is found.
     * 
     * @since 0.9.7
     * @return a list of doublets
     */
    @SuppressWarnings("unchecked")
    protected synchronized List<Class<?>> getDoubletListParallel() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        BlockingQueue<Class<?>> queue = new ArrayBlockingQueue<Class<?>>(
                loadedClassList.size(), false, loadedClassList);
        Class<?> lastClass = loadedClassList.get(loadedClassList.size() - 1);
        if (this.isDoublet(lastClass)) {
            doubletList.add(lastClass);
        }
        // create and start multiple threads
        int n = 2;
        List<Class<?>>[] l = new List[n];
        DoubletDigger[] d = new DoubletDigger[n];
        Thread[] t = new Thread[n];
        for (int i = 0; i < n; i++) {
            l[i] = new ArrayList<Class<?>>();
            d[i] = new DoubletDigger(queue, lastClass, l[i]);
            t[i] = new Thread(d[i]);
            t[i].start();
        }
        // wait for the end of each started thread
        for (int i = 0; i < n; i++) {
            try {
                t[i].join();
            } catch (InterruptedException ie) {
                log.info("interrupted", ie);
            }
        }
        if (log.isTraceEnabled()) {
            log.trace("adding results from " + n + " threads to doubletList");
        }
        for (int i = 0; i < n; i++) {
            doubletList.addAll(l[i]);
        }
        Collections.sort(doubletList, new ObjectComparator());
        return doubletList;
    }
    
    
    /**
     * This class is needed to realize multi threading.
     * @since 0.9.7
     */
    class DoubletDigger implements Runnable {
        
        private final BlockingQueue<Class<?>> queue;
        private final Class<?> lastClass;
        private final List<Class<?>> newDoublets;
        
        public DoubletDigger(BlockingQueue<Class<?>> queue, Class<?> lastClass,
                List<Class<?>> newDoublets) {
            this.queue = queue;
            this.lastClass = lastClass;
            this.newDoublets = newDoublets;
        }

        public void run() {
            if (log.isDebugEnabled()) {
                log.debug("running " + this + "...");
            }
            try {
                while(true) {
                    Class<?> clazz = queue.take();
                    if (clazz == lastClass) {
                        queue.put(clazz);
                        break;
                    }
                    if (doubletList.contains(clazz)) {
                        continue;
                    }
                    try {
                        if (isDoublet(clazz)) {
                            newDoublets.add(clazz);
                        }   
                    } catch (NoSuchElementException nsee) {
                        if (log.isTraceEnabled()) {
                            log.trace(clazz + " not found -> ignored");
                        }
                    }
                }
            } catch (InterruptedException ie) {
                log.info("interrupted", ie);
            }
            if (log.isTraceEnabled()) {
                log.trace(this + "finished, " + newDoublets.size() + " doublets found");
            }
        }
        
    }

}

/*
 * $Log: ClasspathMonitor.java,v $
 * Revision 1.40  2009/08/21 19:59:22  oboehm
 * multi threading support added
 *
 * Revision 1.39  2009/08/11 20:20:23  oboehm
 * 1st try of tuning up ClasspathMonitor by using multiple threads
 *
 * Revision 1.38  2009/08/03 06:12:37  oboehm
 * partially support of WLS classload (via getClasspathFromPackages)
 *
 * Revision 1.37  2009/07/15 21:41:37  oboehm
 * deprecated hint removed
 *
 * Revision 1.36  2009/06/23 19:04:47  oboehm
 * todo added
 *
 * Revision 1.35  2009/05/25 19:10:24  oboehm
 * using now the context classloader to find a resource
 *
 * Revision 1.34  2009/05/20 20:50:14  oboehm
 * logMe() uses now dumpMe(Writer) for logging
 *
 * $Source:
 * /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java
 * /patterntesting/runtime/monitor/ClasspathMonitor.java,v $
 */
