/**
 * $Id: Converter.java,v 1.7 2009/08/03 06:12:37 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 19.01.2009 by oliver (ob@oasd.de)
 */
package patterntesting.runtime.util;

import java.io.File;
import java.net.URI;
import java.util.Set;

import org.apache.commons.lang.StringUtils;

/**
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 19.01.2009
 * @version $Revision: 1.7 $
 */
public class Converter {
    
    public static String getMemoryAsString(long mem) {
        if (mem < 1000) {
            return mem + " bytes";
        } else if (mem < 0x100000) {
            return (mem + 512) / 1024 + " KB";
        } else {
            return (mem + 0x80000) / 0x100000 + " MB";
        }
    }
    
    /**
     * Converts a classname (e.g. "java.lang.String") into a resource
     * ("/java/lang/String.class").).
     * 
     * @param name
     * @return
     */
    public static String classToResource(String name) {
        if (name == null) {
            return null;
        }
        return name.replaceAll("\\.", "/") + ".class";
    }
    
    public static String toResource(Class<?> clazz) {
        return classToResource(clazz.getName());
    }
    
    /**
     * @deprecated use {@link #toResource(Class)}
     */
    public static String classToResource(Class<?> clazz) {
        return classToResource(clazz.getName());
    }
    
    
    /**
     * Converts a package name (e.g. "java.lang") into a resource
     * ("/java/lang").).
     * 
     * @param name
     * @return
     */
    public static String packageToResource(String name) {
        if (name == null) {
            return null;
        }
        return name.replaceAll("\\.", "/");
    }

    public static String toResource(Package p) {
        return packageToResource(p.getName());
    }
    
    /**
     * Converts a resource (e.g. "/java/lang/String.class") into its classname
     * ("java.lang.String").
     * 
     * @param name
     * @return
     */
    public static String resourceToClass(String name) {
        if (name == null) {
            return null;
        }
        if (name.endsWith(".class")) {
            int lastdot = name.lastIndexOf(".");
            return name.substring(0, lastdot).replaceAll("/", "\\.");
        } else {
            return name;
        }
    }
    
    /**
     * Converts an URI into a file.
     * 
     * @param uri e.g. URI("file:/tmp")
     * @return e.g. File("/tmp")
     */
    public static File toFile(URI uri) {
        try {
            return new File(uri);
        } catch (Exception e) {
            String path = uri.getPath();
            if (StringUtils.isEmpty(path)) {
                path = StringUtils.substringAfterLast(uri.toString(), ":");
            }
            return new File(path);
        }
    }
    
    /**
     * Converts an URI into a file and returns it as absolute pathname.
     * 
     * @param uri e.g. URI("file:/tmp")
     * @return e.g. "/tmp"
     */
    public static String toAbsolutePath(URI uri) {
    	File file = toFile(uri);
    	return file.getAbsolutePath();
    }
    
    /**
     * Converts an object into its string representation.
     * A null object is mapped to "" (empty string).
     * 
     * @param obj
     * @return
     */
    public static String toString(Object obj) {
        try {
            if (obj instanceof Object[]) {
                return toString((Object[]) obj);
            }
            return obj.toString();
        } catch (RuntimeException e) {
            return "";
        }
    }
    
    public static String toShortString(Object obj) {
        if (obj == null) {
            return "";
        }
        try {
            Class<?> clazz = (Class<?>) obj;
            return clazz.getSimpleName();
        } catch (ClassCastException cce) {
            return toString(obj);
        }
    }
        
    /**
     * An empty array or null is mapped to "" (empty string).
     * 
     * @param array
     * @return
     */
    public static String toString(Object[] array) {
        try {
            StringBuffer sbuf = new StringBuffer(toString(array[0]));
            for (int i = 1; i < array.length; i++) {
                sbuf.append(", ");
                sbuf.append(toString(array[i]));
            }
            return sbuf.toString();
        } catch (RuntimeException e) {
            return "";
        }
    }
    
    public static String toShortString(Object[] array) {
        try {
            StringBuffer sbuf = new StringBuffer(toShortString(array[0]));
            for (int i = 1; i < array.length; i++) {
                sbuf.append(", ");
                sbuf.append(toShortString(array[i]));
            }
            return sbuf.toString();
        } catch (RuntimeException e) {
            return "";
        }
    }
    
    /**
     * Each object inside the array is converted into its toString()
     * representation.
     * 
     * @since 27-Jul-2009
     * 
     * @param array
     * @return
     */
    public static String[] toStringArray(Object[] array) {
    	String strings[] = new String[array.length];
    	for (int i = 0; i < strings.length; i++) {
			try {
				strings[i] = array[i].toString();
			} catch (NullPointerException npe) {
				strings[i] = null;
			}
		}
    	return strings;
    }

    /**
     * Each object inside the Set is converted into its toString()
     * representation.
     * 
     * @since 27-Jul-2009
     * 
     * @param set
     * @return the given Set as array
     */
	public static String[] toStringArray(Set<? extends Object> set) {
		Object objects[] = new Object[set.size()];
		objects = set.toArray(objects);
		return toStringArray(objects);
	}

}

/**
 * $Log: Converter.java,v $
 * Revision 1.7  2009/08/03 06:12:37  oboehm
 * partially support of WLS classload (via getClasspathFromPackages)
 *
 * Revision 1.6  2009/05/25 19:10:24  oboehm
 * using now the context classloader to find a resource
 *
 * Revision 1.5  2009/02/26 10:02:15  oboehm
 * improved JAMon detection (only JAMon 2.4 will be used if found in classpath)
 * output of ProfileStatistic improved (Label contains no longer return value;
 * this is better for sorting in Excel or OpenOffice)
 *
 * Revision 1.4  2009/02/25 21:22:11  oboehm
 * because of problems with JAMon 2.0 only JAMon 2.4 will be supported
 * (if found in the classpath)
 *
 * Revision 1.3  2009/02/15 21:49:45  oboehm
 * ClasspathChecker rename to ClasspathMonitor
 * getLoadedClasspath() added
 *
 * Revision 1.2  2009/02/11 22:05:54  oboehm
 * query of loaded packages and classes is supported now
 *
 * Revision 1.1  2009/01/19 15:42:57  oboehm
 * MemoryGuard added (to be able to watch memory consumption)
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/util/Converter.java,v $
 */
