/**
 * $Id: ReflectionHelper.java,v 1.4 2009/08/03 06:12:37 oboehm Exp $
 *
 * Copyright (c) 2009 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 09.03.2009 by oliver (ob@aosd.de)
 */
package patterntesting.runtime.util;

import java.lang.reflect.*;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.impl.LogFactoryImpl;

/**
 * This class is a helper class to access some fields via reflection. Normally
 * you should avoid reflection. Handle with care if you use it.
 * 
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 09.03.2009
 * @version $Revision: 1.4 $
 */
public class ReflectionHelper {

     private static final Log log = LogFactoryImpl.getLog(ReflectionHelper.class);

    /**
     * Tries to get the wanted field. Before it is returned the acc
     * 
     * @param cl
     * @param name
     * @return the wanted field
     * @throws NoSuchFieldException
     */
    public static Field getField(Class<?> cl, String name)
            throws NoSuchFieldException {
        try {
            Field field = cl.getDeclaredField(name);
            field.setAccessible(true);
            return field;
        } catch (NoSuchFieldException e) {
            Class<?> superclass = cl.getSuperclass();
            if (superclass == null) {
                throw e;
            } else {
                if (log.isTraceEnabled()) {
                    log.trace("using " + superclass + " to get " + cl.getName()
                            + "." + name + "...");
                }
                return ReflectionHelper.getField(superclass, name);
            }
        }
    }

    /**
     * If it can it returns the value of the given field. If not it throws an
     * exception.
     * 
     * @param obj
     * @param name
     * @return
     * @throws NoSuchFieldException
     * @throws IllegalAccessException
     * @throws IllegalArgumentException
     */
    public static Object getFieldValue(Object obj, String name)
            throws NoSuchFieldException, IllegalArgumentException,
            IllegalAccessException {
        Field field = ReflectionHelper.getField(obj.getClass(), name);
        return field.get(obj);
    }
    
    /**
     * To get all uninitialized field you can call this method.
     * 
     * @param obj the object where to get the fields from
     * @return a collecion of unitialized fields (can be empty)
     */
    public static Collection<Field> getUninitializedNonStaticFields(Object obj) {
        Collection<Field> unitializedFields = new ArrayList<Field>();
        Field[] fields = obj.getClass().getDeclaredFields();
        for (int i = 0; i < fields.length; i++) {
            try {
                fields[i].setAccessible(true);
                if ((fields[i].get(obj) == null) && !isStatic(fields[i])) {
                    unitializedFields.add(fields[i]);
                }
            } catch (IllegalArgumentException e) {
                log.info(e + " => " + fields[i] + " ignored");
            } catch (IllegalAccessException e) {
                log.debug("can't access " + fields[i] + " => ignored");
            }
        }
        return unitializedFields;
    }
    
    private static boolean isStatic(Field field) {
        int m = field.getModifiers();
        return Modifier.isStatic(m);
    }
    
    public static String toShortString(Field field) {
        return field.getType().getSimpleName() + " " + field.getName();
    }
    
    public static String toShortString(Collection<Field> fields) {
        StringBuffer sbuf = new StringBuffer();
        for (Iterator<Field> iterator = fields.iterator(); iterator.hasNext();) {
            Field field = iterator.next();
            sbuf.append(", ");
            sbuf.append(toShortString(field));
        }
        return sbuf.substring(2);
    }

    /**
     * @param cl e.g. class java.lang.ClassLoader
     * @param name e.g. "getPackages"
     * @return a Method object
     * @throws NoSuchMethodException 
     */
	public static Method getMethod(Class<?> cl, String name) throws NoSuchMethodException {
	    try {
	        return cl.getDeclaredMethod(name);
	    } catch (NoSuchMethodException e) {
	        Class<?> superclass = cl.getSuperclass();
	        if (superclass == null) {
	            throw e;
	        } else {
	            return getMethod(superclass, name);
	        }
	    }
	}

}

/**
 * $Log: ReflectionHelper.java,v $
 * Revision 1.4  2009/08/03 06:12:37  oboehm
 * partially support of WLS classload (via getClasspathFromPackages)
 *
 * Revision 1.3  2009/05/08 20:41:08  oboehm
 * https://sourceforge.net/support/tracker.php?aid=2786846 is fixed
 *
 * Revision 1.2  2009/04/13 13:32:48  oboehm
 * some small refactorings
 * Revision 1.1 2009/03/09 21:18:54 oboehm
 * refactored out some parts to ReflectionHelper
 * 
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/util/ReflectionHelper.java,v $
 */
